@@ -24,14 +24,8 @@
 #include <signal.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <ifaddrs.h>
-#include <sys/uio.h>
 #include <sys/poll.h>
 #include <sys/ioctl.h>
-#include <sys/param.h>
-#include <sys/socket.h>
-#include <net/if.h>
-#include <netinet/in.h>
 #include <netinet/tcp.h>
 #include <arpa/inet.h>
 
@@ -39,11 +33,11 @@
 #include "iscsi_proto.h"
 #include "initiator.h"
 #include "iscsi_ipc.h"
-#include "iscsi_sysfs.h"
 #include "log.h"
 #include "transport.h"
-#include "iscsi_settings.h"
 #include "idbm.h"
+#include "iface.h"
+#include "sysdeps.h"
 
 #define LOG_CONN_CLOSED(conn) \
 do { \
@@ -82,6 +76,8 @@ set_non_blocking(int fd)
 
 }
 
+#if 0
+/* not used by anyone */
 static int get_hwaddress_from_netdev(char *netdev, char *hwaddress)
 {
 	struct ifaddrs *ifap, *ifa;
@@ -159,6 +155,7 @@ free_ifap:
 	freeifaddrs(ifap);
 	return found;
 }
+#endif
 
 /*
  * In this mode we do not support interfaces like a bond or alias because
@@ -166,15 +163,14 @@ free_ifap:
  */
 static int get_netdev_from_hwaddress(char *hwaddress, char *netdev)
 {
-	struct ifaddrs *ifap, *ifa;
-	struct sockaddr_in *s4;
-	struct sockaddr_in6 *s6;
+	struct if_nameindex *ifni;
 	struct ifreq if_hwaddr;
-	int found = 0, sockfd;
+	int found = 0, sockfd, i = 0;
 	unsigned char *hwaddr;
-	char buf[INET6_ADDRSTRLEN], tmp_hwaddress[ISCSI_MAX_IFACE_LEN];
+	char tmp_hwaddress[ISCSI_MAX_IFACE_LEN];
 
-	if (getifaddrs(&ifap)) {
+	ifni = if_nameindex();
+	if (ifni == NULL) {
 		log_error("Could not match hwaddress %s to netdev. "
 			  "getifaddrs failed %d", hwaddress, errno);
 		return 0;
@@ -184,34 +180,13 @@ static int get_netdev_from_hwaddress(char *hwaddress, char *netdev)
 	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
 	if (sockfd < 0) {
 		log_error("Could not open socket for ioctl.");
-		goto free_ifap;
+		goto free_ifni;
 	}
 
-	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
-		if (!ifa->ifa_addr)
-			continue;
+	for (i = 0; ifni[i].if_index && ifni[i].if_name; i++) {
+		struct if_nameindex *n = &ifni[i];
 
-		switch (ifa->ifa_addr->sa_family) {
-		case AF_INET:
-			s4 = (struct sockaddr_in *)(ifa->ifa_addr);
-			if (!inet_ntop(ifa->ifa_addr->sa_family,
-				      (void *)&(s4->sin_addr), buf,
-				      INET_ADDRSTRLEN))
-				continue;
-			log_debug(4, "name %s addr %s\n", ifa->ifa_name, buf);
-			break;
-		case AF_INET6:
-			s6 = (struct sockaddr_in6 *)(ifa->ifa_addr);
-			if (!inet_ntop(ifa->ifa_addr->sa_family,
-			    (void *)&(s6->sin6_addr), buf, INET6_ADDRSTRLEN))
-				continue;
-			log_debug(4, "name %s addr %s\n", ifa->ifa_name, buf);
-			break;
-		default:
-			continue;
-		}
-
-		strncpy(if_hwaddr.ifr_name, ifa->ifa_name, IFNAMSIZ);
+		strlcpy(if_hwaddr.ifr_name, n->if_name, IFNAMSIZ);
 		if (ioctl(sockfd, SIOCGIFHWADDR, &if_hwaddr) < 0) {
 			log_error("Could not match %s to netdevice.",
 				  hwaddress);
@@ -231,20 +206,27 @@ static int get_netdev_from_hwaddress(char *hwaddress, char *netdev)
 		log_debug(4, "Found hardware address %s", tmp_hwaddress);
 		if (!strcasecmp(tmp_hwaddress, hwaddress)) {
 			log_debug(4, "Matches %s to %s", hwaddress,
-				  ifa->ifa_name);
+				  n->if_name);
 			memset(netdev, 0, IFNAMSIZ); 
-			strncpy(netdev, ifa->ifa_name, IFNAMSIZ);
+			strlcpy(netdev, n->if_name, IFNAMSIZ);
 			found = 1;
 			break;
 		}
 	}
 
 	close(sockfd);
-free_ifap:
-	freeifaddrs(ifap);
+free_ifni:
+	if_freenameindex(ifni);
 	return found;
 }
 
+#if 0
+
+This is not supported for now, because it is not exactly what we want.
+It also turns out that targets will send packets to other interfaces
+causing all types of weird things to happen.
+
+
 static int bind_src_by_address(int sockfd, char *address)
 {
 	int rc = 0;
@@ -275,6 +257,7 @@ static int bind_src_by_address(int sockfd, char *address)
 		log_debug(4, "Bound %s to socket fd %d", address, sockfd);
 	return rc;
 }
+#endif
 
 static int bind_conn_to_iface(iscsi_conn_t *conn, struct iface_rec *iface)
 {
@@ -286,13 +269,20 @@ static int bind_conn_to_iface(iscsi_conn_t *conn, struct iface_rec *iface)
 		log_error("Cannot match %s to net/scsi interface.",
 			  iface->hwaddress);
                 return -1;
-	} else if (iface_is_bound_by_ipaddr(iface) &&
-		   bind_src_by_address(conn->socket_fd, iface->ipaddress)) {
-		log_error("Cannot match %s to net/scsi interface.",
-			   iface->ipaddress);
-		return -1;
 	} else if (iface_is_bound_by_netdev(iface))
 		strcpy(session->netdev, iface->netdev);
+	else if (iface_is_bound_by_ipaddr(iface)) {
+		/*
+		 * we never supported this but now with offload having to
+		 * set the ip address in the iface, useris may forget to
+		 * set the offload's transport type and we end up here by
+		 * accident.
+		 */
+		log_error("Cannot bind %s to net/scsi interface. This is not "
+			  "supported with software iSCSI (iscsi_tcp).",
+			   iface->ipaddress);
+		return -1;
+	}
 
 	if (strlen(session->netdev)) {
 		struct ifreq ifr;
@@ -300,7 +290,7 @@ static int bind_conn_to_iface(iscsi_conn_t *conn, struct iface_rec *iface)
 		log_debug(4, "Binding session %d to %s", session->id,
 			  session->netdev);
 		memset(&ifr, 0, sizeof(ifr));
-		strncpy(ifr.ifr_name, session->netdev, IFNAMSIZ);
+		strlcpy(ifr.ifr_name, session->netdev, IFNAMSIZ);
 
 		if (setsockopt(conn->socket_fd, SOL_SOCKET, SO_BINDTODEVICE,
 			       session->netdev,
@@ -417,8 +407,8 @@ iscsi_io_tcp_poll(iscsi_conn_t *conn, int timeout_ms)
 			    conn->host, sizeof(conn->host), serv, sizeof(serv),
 			    NI_NUMERICHOST|NI_NUMERICSERV);
 
-		log_error("cannot make connection to %s:%s (%d)",
-			  conn->host, serv, errno);
+		log_error("cannot make connection to %s:%s (%s)",
+			  conn->host, serv, strerror(errno));
 		return rc;
 	}
 
@@ -429,7 +419,13 @@ iscsi_io_tcp_poll(iscsi_conn_t *conn, int timeout_ms)
 		return -1;
 	}
 	if (rc) {
-		log_error("connect failed (%d)\n", rc);
+		getnameinfo((struct sockaddr *) &conn->saddr,
+			    sizeof(conn->saddr),
+			    conn->host, sizeof(conn->host), serv, sizeof(serv),
+			    NI_NUMERICHOST|NI_NUMERICSERV);
+
+		log_error("connect to %s:%s failed (%s)\n",
+			  conn->host, serv, strerror(rc));
 		return -rc;
 	}
 
@@ -569,7 +565,7 @@ iscsi_io_send_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	/* set a timeout, since the socket calls may take a long time
 	 * to timeout on their own
 	 */
-	if (!conn->kernel_io) {
+	if (!ipc) {
 		memset(&action, 0, sizeof (struct sigaction));
 		memset(&old, 0, sizeof (struct sigaction));
 		action.sa_sigaction = NULL;
@@ -627,23 +623,21 @@ iscsi_io_send_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	end = header + sizeof (*hdr) + hdr->hlength;
 
 	/* send all the data and any padding */
-	if (pdu_length % PAD_WORD_LEN)
-		pad_bytes = PAD_WORD_LEN - (pdu_length % PAD_WORD_LEN);
+	if (pdu_length % ISCSI_PAD_LEN)
+		pad_bytes = ISCSI_PAD_LEN - (pdu_length % ISCSI_PAD_LEN);
 	else
 		pad_bytes = 0;
 
-	if (conn->kernel_io) {
-		conn->send_pdu_begin(session->t->handle, session->id,
-			conn->id, end - header,
-			ntoh24(hdr->dlength) + pad_bytes);
-		conn->send_pdu_timer_add(conn, timeout);
-	}
+	if (ipc)
+		ipc->send_pdu_begin(session->t->handle, session->id,
+				    conn->id, end - header,
+				    ntoh24(hdr->dlength) + pad_bytes);
 
 	while (header < end) {
 		vec[0].iov_base = header;
 		vec[0].iov_len = end - header;
 
-		if (!conn->kernel_io)
+		if (!ipc)
 			rc = writev(session->ctrl_fd, vec, 1);
 		else
 			rc = ipc->writev(0, vec, 1);
@@ -671,7 +665,7 @@ iscsi_io_send_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 		vec[1].iov_base = (void *) &pad;
 		vec[1].iov_len = pad_bytes;
 
-		if (!conn->kernel_io)
+		if (!ipc)
 			rc = writev(session->ctrl_fd, vec, 2);
 		else
 			rc = ipc->writev(0, vec, 2);
@@ -695,10 +689,9 @@ iscsi_io_send_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 		}
 	}
 
-	if (conn->kernel_io) {
-		if (conn->send_pdu_end(session->t->handle, session->id,
-				       conn->id, &rc)) {
-			conn->send_pdu_timer_remove(conn);
+	if (ipc) {
+		if (ipc->send_pdu_end(session->t->handle, session->id,
+				      conn->id, &rc)) {
 			ret = 0;
 			goto done;
 		}
@@ -707,7 +700,7 @@ iscsi_io_send_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	ret = 1;
 
       done:
-	if (!conn->kernel_io) {
+	if (!ipc) {
 		alarm(0);
 		sigaction(SIGALRM, &old, NULL);
 		timedout = 0;
@@ -732,8 +725,6 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	char *end = data + max_data_length;
 	struct sigaction action;
 	struct sigaction old;
-	uintptr_t pdu_handle;
-	int pdu_size;
 	iscsi_session_t *session = conn->session;
 
 	memset(data, 0, max_data_length);
@@ -741,7 +732,7 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	/* set a timeout, since the socket calls may take a long
 	 * time to timeout on their own
 	 */
-	if (!conn->kernel_io) {
+	if (!ipc) {
 		memset(&action, 0, sizeof (struct sigaction));
 		memset(&old, 0, sizeof (struct sigaction));
 		action.sa_sigaction = NULL;
@@ -751,8 +742,7 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 		timedout = 0;
 		alarm(timeout);
 	} else {
-		if (conn->recv_pdu_begin(session->ctrl_fd,
-				conn->recv_handle, &pdu_handle, &pdu_size)) {
+		if (ipc->recv_pdu_begin(conn)) {
 			failed = 1;
 			goto done;
 		}
@@ -760,7 +750,7 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 
 	/* read a response header */
 	do {
-		if (!conn->kernel_io)
+		if (!ipc)
 			rlen = read(session->ctrl_fd, header,
 					sizeof (*hdr) - h_bytes);
 		else
@@ -807,7 +797,7 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	if (dlength == 0)
 		goto done;
 
-	if (data + dlength >= end) {
+	if (data + dlength > end) {
 		log_warning("buffer size %u too small for data length %u",
 		       max_data_length, dlength);
 		failed = 1;
@@ -817,7 +807,7 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	/* read the rest into our buffer */
 	d_bytes = 0;
 	while (d_bytes < dlength) {
-		if (!conn->kernel_io)
+		if (!ipc)
 			rlen = read(session->ctrl_fd, data + d_bytes,
 					dlength - d_bytes);
 		else
@@ -843,10 +833,10 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 
 	/* handle PDU data padding.
 	 * data is padded in case of kernel_io */
-	pad = dlength % PAD_WORD_LEN;
-	if (pad && !conn->kernel_io) {
-		int pad_bytes = pad = PAD_WORD_LEN - pad;
-		char bytes[PAD_WORD_LEN];
+	pad = dlength % ISCSI_PAD_LEN;
+	if (pad && !ipc) {
+		int pad_bytes = pad = ISCSI_PAD_LEN - pad;
+		char bytes[ISCSI_PAD_LEN];
 
 		while (pad_bytes > 0) {
 			rlen = read(conn->socket_fd, &bytes, pad_bytes);
@@ -900,16 +890,14 @@ iscsi_io_recv_pdu(iscsi_conn_t *conn, struct iscsi_hdr *hdr,
 	}
 
 done:
-	if (!conn->kernel_io) {
+	if (!ipc) {
 		alarm(0);
 		sigaction(SIGALRM, &old, NULL);
 	} else {
 		/* finalyze receive transaction */
-		if (conn->recv_pdu_end(session->ctrl_fd, (uintptr_t)conn,
-				pdu_handle)) {
+		if (ipc->recv_pdu_end(conn)) {
 			failed = 1;
 		}
-		conn->send_pdu_timer_remove(conn);
 	}
 
 	if (timedout || failed) {
