@@ -41,6 +41,9 @@
 #include "iscsi_settings.h"
 #include "fw_context.h"
 #include "iface.h"
+#include "session_info.h"
+#include "host.h"
+#include "sysdeps.h"
 
 struct iscsi_ipc *ipc = NULL; /* dummy */
 static char program_name[] = "iscsiadm";
@@ -73,6 +76,7 @@ static struct option const long_options[] =
 	{"type", required_argument, NULL, 't'},
 	{"name", required_argument, NULL, 'n'},
 	{"value", required_argument, NULL, 'v'},
+	{"host", required_argument, NULL, 'H'},
 	{"sid", required_argument, NULL, 'r'},
 	{"rescan", no_argument, NULL, 'R'},
 	{"print", required_argument, NULL, 'P'},
@@ -104,6 +108,7 @@ iscsiadm -m node [ -hV ] [ -d debug_level ] [ -P printlevel ] [ -L all,manual,au
 iscsiadm -m session [ -hV ] [ -d debug_level ] [ -P  printlevel] [ -r sessionid | sysfsdir [ -R | -u | -s ] [ -o operation ] [ -n name ] [ -v value ] ]\n\
 iscsiadm -m iface [ -hV ] [ -d debug_level ] [ -P printlevel ] [ -I ifacename ] [ [ -o  operation  ] [ -n name ] [ -v value ] ]\n\
 iscsiadm -m fw [ -l ]\n\
+iscsiadm -m host [ -P printlevel ] [ -H hostno ]\n\
 iscsiadm -k priority\n");
 	}
 	exit(status == 0 ? 0 : -1);
@@ -143,6 +148,8 @@ str_to_mode(char *str)
 		mode = MODE_IFACE;
 	else if (!strcmp("fw", str))
 		mode = MODE_FW;
+	else if (!strcmp("host", str))
+		mode = MODE_HOST;
 	else
 		mode = -1;
 
@@ -162,7 +169,7 @@ str_to_type(char *str)
 	else if (!strcmp("isns", str))
 		type = DISCOVERY_TYPE_ISNS;
 	else if (!strcmp("fw", str))
-		type = DISCOVERY_TYPE_FWBOOT;
+		type = DISCOVERY_TYPE_FW;
 	else
 		type = -1;
 
@@ -206,7 +213,7 @@ static void kill_iscsid(int priority)
  * (scsi_host and iscsi_sessions are the currently running instance of
  * a iface or node record).
  */
-static int print_ifaces(int info_level)
+static int print_ifaces(struct iface_rec *iface, int info_level)
 {
 	int err, num_found = 0;
 
@@ -217,11 +224,21 @@ static int print_ifaces(int info_level)
 					   iface_print_flat);
 		break;
 	case 1:
-		err = iface_for_each_iface(NULL, &num_found,
-					   iface_print_tree);
+		if (iface) {
+			err = iface_conf_read(iface);
+			if (err) {
+				log_error("Could not read iface %s.\n",
+					  iface->name);
+				return err;
+			}
+			iface_print_tree(NULL, iface);
+			num_found = 1;
+		} else
+			err = iface_for_each_iface(NULL, &num_found,
+						   iface_print_tree);
 		break;
 	default:
-		log_error("Invalid info level %d. Try 0 or 1.", info_level);
+		log_error("Invalid info level %d. Try 0 - 1.", info_level);
 		return EINVAL;
 	}
 
@@ -340,50 +357,6 @@ for_each_session(struct node_rec *rec, iscsi_sysfs_session_op_fn *fn)
 	return err;
 }
 
-struct session_link_info {
-	struct list_head *list;
-	struct session_info *match_info;
-};
-
-static int link_sessions(void *data, struct session_info *info)
-{
-	struct session_link_info *link_info = data;
-	struct list_head *list = link_info->list;
-	struct session_info *new, *curr, *match = NULL;
-
-	if (link_info->match_info && link_info->match_info->sid != info->sid)
-		return 0;
-
-	new = calloc(1, sizeof(*new));
-	if (!new)
-		return ENOMEM;
-	memcpy(new, info, sizeof(*new));
-	INIT_LIST_HEAD(&new->list);
-
-	if (list_empty(list)) {
-		list_add_tail(&new->list, list);
-		return 0;
-	}
-
-	list_for_each_entry(curr, list, list) {
-		if (!strcmp(curr->targetname, info->targetname)) {
-			match = curr;
-
-			if (!strcmp(curr->address, info->address)) {
-				match = curr;
-
-				if (curr->port == info->port) {
-					match = curr;
-					break;
-				}
-			}
-		}
-	}
-
-	list_add_tail(&new->list, match ? match->list.next : list);
-	return 0;
-}
-
 static int link_recs(void *data, struct node_rec *rec)
 {
 	struct list_head *list = data;
@@ -447,7 +420,7 @@ __logout_portals(void *data, int *nr_found,
 		 int (*logout_fn)(void *, struct list_head *,
 				   struct session_info *))
 {
-	struct session_info *curr_info, *tmp;
+	struct session_info *curr_info;
 	struct session_link_info link_info;
 	struct list_head session_list, logout_list;
 	int ret = 0, err;
@@ -457,11 +430,12 @@ __logout_portals(void *data, int *nr_found,
 
 	memset(&link_info, 0, sizeof(link_info));
 	link_info.list = &session_list;
-	link_info.match_info = NULL;
+	link_info.data = NULL;
+	link_info.match_fn = NULL;
 	*nr_found = 0;
 
 	err = iscsi_sysfs_for_each_session(&link_info, nr_found,
-				    link_sessions);
+					   session_info_create_list);
 	if (err || !*nr_found)
 		return err;
 
@@ -478,10 +452,7 @@ __logout_portals(void *data, int *nr_found,
 	if (err)
 		ret = err;
 
-	list_for_each_entry_safe(curr_info, tmp, &session_list, list) {
-		list_del(&curr_info->list);
-		free(curr_info);
-	}
+	session_info_free_list(&session_list);
 	return ret;
 }
 
@@ -584,11 +555,11 @@ create_node_record(char *targetname, int tpgt, char *ip, int port,
 
 	idbm_node_setup_defaults(rec);
 	if (targetname)
-		strncpy(rec->name, targetname, TARGET_NAME_MAXLEN);
+		strlcpy(rec->name, targetname, TARGET_NAME_MAXLEN);
 	rec->tpgt = tpgt;
 	rec->conn[0].port = port;
 	if (ip)
-		strncpy(rec->conn[0].address, ip, NI_MAXHOST);
+		strlcpy(rec->conn[0].address, ip, NI_MAXHOST);
 	memset(&rec->iface, 0, sizeof(struct iface_rec));
 	if (iface) {
 		iface_copy(&rec->iface, iface);
@@ -803,7 +774,7 @@ static int print_nodes(int info_level, struct node_rec *rec)
 		break;
 	case 1:
 		memset(&tmp_rec, 0, sizeof(node_rec_t));
-		if (for_each_rec(rec, &tmp_rec, idbm_print_node_tree))
+		if (for_each_rec(rec, &tmp_rec, idbm_print_node_and_iface_tree))
 			rc = -1;
 		break;
 	default:
@@ -835,284 +806,6 @@ static char *get_config_file(void)
 	return NULL;
 }
 
-static int print_session_flat(void *data, struct session_info *info)
-{
-	struct session_info *match_info = data;
-	struct iscsi_transport *t = iscsi_sysfs_get_transport_by_sid(info->sid);
-
-	if (match_info && match_info->sid != info->sid)
-		return 0;
-
-	if (strchr(info->persistent_address, '.'))
-		printf("%s: [%d] %s:%d,%d %s\n",
-			t ? t->name : UNKNOWN_VALUE,
-			info->sid, info->persistent_address,
-			info->persistent_port, info->tpgt, info->targetname);
-	else
-		printf("%s: [%d] [%s]:%d,%d %s\n",
-			t ? t->name : UNKNOWN_VALUE,
-			info->sid, info->persistent_address,
-			info->persistent_port, info->tpgt, info->targetname);
-	return 0;
-}
-
-static int print_iscsi_state(int sid)
-{
-	iscsiadm_req_t req;
-	iscsiadm_rsp_t rsp;
-	int err;
-	char *state = NULL;
-	char state_buff[SCSI_MAX_STATE_VALUE];
-	static char *conn_state[] = {
-		"FREE",
-		"TRANSPORT WAIT",
-		"IN LOGIN",
-		"LOGGED IN",
-		"IN LOGOUT",
-		"LOGOUT REQUESTED",
-		"CLEANUP WAIT",
-	};
-	static char *session_state[] = {
-		"NO CHANGE",
-		"CLEANUP",
-		"REPOEN",
-		"REDIRECT",
-	};
-
-	memset(&req, 0, sizeof(iscsiadm_req_t));
-	req.command = MGMT_IPC_SESSION_INFO;
-	req.u.session.sid = sid;
-
-	err = do_iscsid(&req, &rsp);
-	/*
-	 * for drivers like qla4xxx, iscsid does not display
-	 * anything here since it does not know about it.
-	 */
-	if (!err && rsp.u.session_state.conn_state >= 0 &&
-	    rsp.u.session_state.conn_state <= STATE_CLEANUP_WAIT)
-		state = conn_state[rsp.u.session_state.conn_state];
-	printf("\t\tiSCSI Connection State: %s\n", state ? state : "Unknown");
-	state = NULL;
-
-	memset(state_buff, 0, SCSI_MAX_STATE_VALUE);
-	if (!iscsi_sysfs_get_session_state(state_buff, sid))
-		printf("\t\tiSCSI Session State: %s\n", state_buff);
-	else
-		printf("\t\tiSCSI Session State: Unknown\n");
-
-	if (!err && rsp.u.session_state.session_state >= 0 &&
-	   rsp.u.session_state.session_state <= R_STAGE_SESSION_REDIRECT)
-		state = session_state[rsp.u.session_state.session_state];
-	printf("\t\tInternal iscsid Session State: %s\n",
-	       state ? state : "Unknown");
-	return 0;
-}
-
-static void print_iscsi_params(int sid)
-{
-	struct iscsi_session_operational_config session_conf;
-	struct iscsi_conn_operational_config conn_conf;
-
-	iscsi_sysfs_get_negotiated_session_conf(sid, &session_conf);
-	iscsi_sysfs_get_negotiated_conn_conf(sid, &conn_conf);
-
-	printf("\t\t************************\n");
-	printf("\t\tNegotiated iSCSI params:\n");
-	printf("\t\t************************\n");
-
-	if (is_valid_operational_value(conn_conf.HeaderDigest))
-		printf("\t\tHeaderDigest: %s\n",
-			conn_conf.HeaderDigest ? "CRC32C" : "None");
-	if (is_valid_operational_value(conn_conf.DataDigest))
-		printf("\t\tDataDigest: %s\n",
-			conn_conf.DataDigest ? "CRC32C" : "None");
-	if (is_valid_operational_value(conn_conf.MaxRecvDataSegmentLength))
-		printf("\t\tMaxRecvDataSegmentLength: %d\n",
-			conn_conf.MaxRecvDataSegmentLength);
-	if (is_valid_operational_value(conn_conf.MaxXmitDataSegmentLength))
-		printf("\t\tMaxXmitDataSegmentLength: %d\n",
-			conn_conf.MaxXmitDataSegmentLength);
-	if (is_valid_operational_value(session_conf.FirstBurstLength))
-		printf("\t\tFirstBurstLength: %d\n",
-			session_conf.FirstBurstLength);
-	if (is_valid_operational_value(session_conf.MaxBurstLength))
-		printf("\t\tMaxBurstLength: %d\n",
-			session_conf.MaxBurstLength);
-	if (is_valid_operational_value(session_conf.ImmediateData))
-		printf("\t\tImmediateData: %s\n",
-			session_conf.ImmediateData ? "Yes" : "No");
-	if (is_valid_operational_value(session_conf.InitialR2T))
-		printf("\t\tInitialR2T: %s\n",
-			session_conf.InitialR2T ? "Yes" : "No");
-	if (is_valid_operational_value(session_conf.MaxOutstandingR2T))
-		printf("\t\tMaxOutstandingR2T: %d\n",
-			session_conf.MaxOutstandingR2T);
-}
-
-static void print_scsi_device_info(int host_no, int target, int lun)
-{
-	char *blockdev, state[SCSI_MAX_STATE_VALUE];
-
-	printf("\t\tscsi%d Channel 00 Id %d Lun: %d\n", host_no, target, lun);
-	blockdev = iscsi_sysfs_get_blockdev_from_lun(host_no, target, lun);
-	if (blockdev) {
-		printf("\t\t\tAttached scsi disk %s\t\t", blockdev);
-		free(blockdev);
-
-		if (!iscsi_sysfs_get_device_state(state, host_no, target, lun))
-			printf("State: %s\n", state);
-		else
-			printf("State: Unknown\n");
-	}
-}
-
-static int print_scsi_state(int sid)
-{
-	int host_no = -1, err = 0;
-	char state[SCSI_MAX_STATE_VALUE];
-
-	printf("\t\t************************\n");
-	printf("\t\tAttached SCSI devices:\n");
-	printf("\t\t************************\n");
-
-	host_no = iscsi_sysfs_get_host_no_from_sid(sid, &err);
-	if (err) {
-		printf("\t\tHost No: Unknown\n");
-		return err;
-	}
-	printf("\t\tHost Number: %d\t", host_no);
-	if (!iscsi_sysfs_get_host_state(state, host_no))
-		printf("State: %s\n", state);
-	else
-		printf("State: Unknown\n");
-
-	iscsi_sysfs_for_each_device(host_no, sid, print_scsi_device_info);
-	return 0;
-}
-
-static void print_sessions_tree(struct list_head *list, int level)
-{
-	struct session_info *curr, *prev = NULL, *tmp;
-	struct iscsi_transport *t;
-
-	list_for_each_entry(curr, list, list) {
-		if (!prev || strcmp(prev->targetname, curr->targetname)) {
-			printf("Target: %s\n", curr->targetname);
-			prev = NULL;
-		}
-
-		if (!prev || (strcmp(prev->address, curr->address) ||
-		     prev->port != curr->port)) {
-			if (strchr(curr->address, '.'))
-				printf("\tCurrent Portal: %s:%d,%d\n",
-				      curr->address, curr->port, curr->tpgt);
-			else
-				printf("\tCurrent Portal: [%s]:%d,%d\n",
-				      curr->address, curr->port, curr->tpgt);
-
-			if (strchr(curr->persistent_address, '.'))
-				printf("\tPersistent Portal: %s:%d,%d\n",
-				      curr->persistent_address,
-				      curr->persistent_port, curr->tpgt);
-			else
-				printf("\tPersistent Portal: [%s]:%d,%d\n",
-				      curr->persistent_address,
-				      curr->persistent_port, curr->tpgt);
-		} else
-			printf("\n");
-
-		t = iscsi_sysfs_get_transport_by_sid(curr->sid);
-
-		printf("\t\t**********\n");
-		printf("\t\tInterface:\n");
-		printf("\t\t**********\n");
-		if (strlen(curr->iface.name))
-			printf("\t\tIface Name: %s\n", curr->iface.name);
-		else
-			printf("\t\tIface Name: %s\n", UNKNOWN_VALUE);
-		printf("\t\tIface Transport: %s\n",
-		       t ? t->name : UNKNOWN_VALUE);
-		printf("\t\tIface Initiatorname: %s\n",
-		       strlen(curr->iface.iname) ? curr->iface.iname :
-								UNKNOWN_VALUE);
-		if (strchr(curr->address, '.'))
-			printf("\t\tIface IPaddress: %s\n",
-			       curr->iface.ipaddress);
-		else
-			printf("\t\tIface IPaddress: [%s]\n",
-			       curr->iface.ipaddress);
-		printf("\t\tIface HWaddress: %s\n", curr->iface.hwaddress);
-		printf("\t\tIface Netdev: %s\n", curr->iface.netdev);
-		printf("\t\tSID: %d\n", curr->sid);
-		print_iscsi_state(curr->sid);
-
-		if (level < 2)
-			goto next;
-		print_iscsi_params(curr->sid);
-
-		if (level < 3)
-			goto next;
-		print_scsi_state(curr->sid);
-next:
-		prev = curr;
-	}
-
-	list_for_each_entry_safe(curr, tmp, list, list) {
-		list_del(&curr->list);
-		free(curr);
-	}
-}
-
-static int print_sessions(int info_level, struct session_info *match_info)
-{
-	struct list_head list;
-	int num_found = 0, err = 0;
-	char *version;
-
-	switch (info_level) {
-	case 0:
-	case -1:
-		err = iscsi_sysfs_for_each_session(match_info, &num_found,
-						   print_session_flat);
-		break;
-	case 2:
-	case 3:
-		version = iscsi_sysfs_get_iscsi_kernel_version();
-		if (version) {
-			printf("iSCSI Transport Class version %s\n",
-				version);
-			printf("%s version %s\n", program_name,
-			      ISCSI_VERSION_STR);
-		}
-		/* fall through */
-	case 1:
-		INIT_LIST_HEAD(&list);
-		struct session_link_info link_info;
-
-		memset(&link_info, 0, sizeof(link_info));
-		link_info.list = &list;
-		link_info.match_info = match_info;
-
-		err = iscsi_sysfs_for_each_session(&link_info, &num_found,
-						   link_sessions);
-		if (err || !num_found)
-			break;
-
-		print_sessions_tree(&list, info_level);
-		break;
-	default:
-		log_error("Invalid info level %d. Try 0 - 3.", info_level);
-		return EINVAL;
-	}
-
-	if (err) {
-		log_error("Can not get list of active sessions (%d)", err);
-		return err;
-	} else if (!num_found)
-		log_error("No active sessions.");
-	return 0;
-}
-
 static int rescan_portal(void *data, struct session_info *info)
 {
 	int host_no, err;
@@ -1130,7 +823,7 @@ static int rescan_portal(void *data, struct session_info *info)
 		return err;
 	}
 	/* rescan each device to pick up size changes */
-	iscsi_sysfs_for_each_device(host_no, info->sid,
+	iscsi_sysfs_for_each_device(NULL, host_no, info->sid,
 				    iscsi_sysfs_rescan_device);
 	/* now scan for new devices */
 	iscsi_sysfs_scan_host(host_no, 0);
@@ -1245,10 +938,10 @@ static int add_static_rec(int *found, char *targetname, int tpgt,
 	drec->type = DISCOVERY_TYPE_STATIC;
 
 	idbm_node_setup_from_conf(rec);
-	strncpy(rec->name, targetname, TARGET_NAME_MAXLEN);
+	strlcpy(rec->name, targetname, TARGET_NAME_MAXLEN);
 	rec->tpgt = tpgt;
 	rec->conn[0].port = port;
-	strncpy(rec->conn[0].address, ip, NI_MAXHOST);
+	strlcpy(rec->conn[0].address, ip, NI_MAXHOST);
 
 	if (iface) {
 		rc = iface_conf_read(iface);
@@ -1400,7 +1093,7 @@ static int delete_node(void *data, struct node_rec *rec)
 	return idbm_delete_node(rec);
 }
 
-static int delete_stale_recs(void *data, struct node_rec *rec)
+static int delete_stale_rec(void *data, struct node_rec *rec)
 {
 	struct list_head *new_rec_list = data;
 	struct node_rec *new_rec;
@@ -1432,33 +1125,19 @@ static int delete_stale_recs(void *data, struct node_rec *rec)
 }
 
 static int
-update_discovery_recs(discovery_rec_t *drec,
-		      struct list_head *new_rec_list, struct list_head *ifaces,
-		      int info_level, int do_login, int op)
+exec_disc_op_on_recs(discovery_rec_t *drec, struct list_head *rec_list,
+		     int info_level, int do_login, int op)
 {
-	int rc, err, found = 0;
-	struct list_head bound_rec_list;
-	struct node_rec *new_rec, *tmp;
-
-	INIT_LIST_HEAD(&bound_rec_list);
- 
-	/* bind ifaces to node recs so we know what we have */
-	list_for_each_entry(new_rec, new_rec_list, list) {
-		rc = idbm_bind_ifaces_to_node(new_rec, ifaces,
-					      &bound_rec_list);
-		if (rc)
-			goto free_bound_recs;
-	}
-
+	int rc = 0, err, found = 0;
+	struct node_rec *new_rec;
 
 	/* clean up node db */
 	if (op & OP_DELETE)
-		idbm_for_each_rec(&found, &bound_rec_list,
-				  delete_stale_recs);
+		idbm_for_each_rec(&found, rec_list, delete_stale_rec);
 
 	if (op & OP_NEW || op & OP_UPDATE) {
 		/* now add/update records */
-		list_for_each_entry(new_rec, &bound_rec_list, list) {
+		list_for_each_entry(new_rec, rec_list, list) {
 			rc = idbm_add_node(new_rec, drec, op & OP_UPDATE);
 			if (rc)
 				log_error("Could not add/update "
@@ -1473,34 +1152,24 @@ update_discovery_recs(discovery_rec_t *drec,
 
 	idbm_print_discovered(drec, info_level);
 
-	if (!do_login) {
-		rc = 0;
-		goto free_bound_recs;
-	}
+	if (!do_login)
+		return 0;
 
-	err = __login_portals(drec, &found, &bound_rec_list,
-			      login_discovered_portal);
+	err = __login_portals(drec, &found, rec_list, login_discovered_portal);
 	if (err && !rc)
 		rc = err;
-
-free_bound_recs:
-	list_for_each_entry_safe(new_rec, tmp, &bound_rec_list, list) {
-		list_del(&new_rec->list);
-		free(new_rec);
-	}
 	return rc;
 }
 
 static int
-do_sofware_sendtargets(discovery_rec_t *drec, struct list_head *ifaces,
-		       int info_level, int do_login,
-			int op)
+do_software_sendtargets(discovery_rec_t *drec, struct list_head *ifaces,
+		        int info_level, int do_login, int op)
 {
-	struct list_head new_rec_list;
-	struct node_rec *new_rec, *tmp;
+	struct list_head rec_list;
+	struct node_rec *rec, *tmp;
 	int rc;
 
-	INIT_LIST_HEAD(&new_rec_list);
+	INIT_LIST_HEAD(&rec_list);
 	/*
 	 * compat: if the user did not pass any op then we do all
 	 * ops for them
@@ -1509,24 +1178,31 @@ do_sofware_sendtargets(discovery_rec_t *drec, struct list_head *ifaces,
 		op = OP_NEW | OP_DELETE | OP_UPDATE;
 
 	drec->type = DISCOVERY_TYPE_SENDTARGETS;
-	rc = discovery_sendtargets(drec, &new_rec_list);
-	if (rc)
-		return rc;
-
+	/*
+	 * we will probably want to know how a specific iface and discovery
+	 * DB lined up, but for now just put all the targets found from
+	 * a discovery portal in one place
+	 */
 	rc = idbm_add_discovery(drec, op & OP_UPDATE);
 	if (rc) {
 		log_error("Could not add new discovery record.");
-		goto free_new_recs;
+		return rc;
 	}
 
-	rc = update_discovery_recs(drec, &new_rec_list, ifaces,
-				   info_level, do_login, op);
-
-free_new_recs:
-	list_for_each_entry_safe(new_rec, tmp, &new_rec_list, list) {
-		list_del(&new_rec->list);
-		free(new_rec);
+	rc = idbm_bind_ifaces_to_nodes(discovery_sendtargets, drec, ifaces,
+				       &rec_list);
+	if (rc) {
+		log_error("Could not perform SendTargets discovery.");
+		return rc;
 	}
+
+	rc = exec_disc_op_on_recs(drec, &rec_list, info_level, do_login, op);
+
+	list_for_each_entry_safe(rec, tmp, &rec_list, list) {
+		list_del(&rec->list);
+		free(rec);
+	}
+
 	return rc;
 }
 
@@ -1556,7 +1232,7 @@ do_sendtargets(discovery_rec_t *drec, struct list_head *ifaces,
 			continue;
 		}
 
-		host_no = iscsi_sysfs_get_host_no_from_iface(iface, &rc);
+		host_no = iscsi_sysfs_get_host_no_from_hwinfo(iface, &rc);
 		if (rc || host_no == -1) {
 			log_debug(1, "Could not match iface" iface_fmt " to "
 				  "host.", iface_str(iface)); 
@@ -1587,24 +1263,8 @@ do_sendtargets(discovery_rec_t *drec, struct list_head *ifaces,
 		return ENODEV;
 
 sw_st:
-	return do_sofware_sendtargets(drec, ifaces, info_level, do_login,
-				      op);
-}
-
-/* TODO: merge this with the idbm code */
-static void print_fw_nodes(struct node_rec *rec, int info_level)
-{
-	switch (info_level) {
-	case -1:
-	case 0:
-		idbm_print_node_flat(NULL, rec);
-		break;
-	case 1:
-		idbm_print_node_tree(NULL, rec);
-		break;
-	default:
-		log_error("Invalid print level %d. Try 0 or 1.", info_level);
-	}
+	return do_software_sendtargets(drec, ifaces, info_level, do_login,
+				       op);
 }
 
 static int isns_dev_attr_query(discovery_rec_t *drec,
@@ -1778,9 +1438,9 @@ update_fail:
 			  iface->name);
 		break;
 	default:
-		if (!iface) {
+		if (!iface || (iface && info_level > 0)) {
 			if (op == OP_NOOP || op == OP_SHOW)
-				rc = print_ifaces(info_level);
+				rc = print_ifaces(iface, info_level);
 			else
 				rc = EINVAL;
 		} else {
@@ -1924,8 +1584,7 @@ out:
 	return rc;
 }
 
-static struct node_rec *
-fw_create_rec_by_entry(struct boot_context *context)
+struct node_rec *fw_create_rec_by_entry(struct boot_context *context)
 {
 	struct node_rec *rec;
 
@@ -1940,15 +1599,15 @@ fw_create_rec_by_entry(struct boot_context *context)
 
 	/* todo - grab mac and set that here */
 	iface_setup_defaults(&rec->iface);
-	strncpy(rec->iface.iname, context->initiatorname,
+	strlcpy(rec->iface.iname, context->initiatorname,
 		sizeof(context->initiatorname));
-	strncpy(rec->session.auth.username, context->chap_name,
+	strlcpy(rec->session.auth.username, context->chap_name,
 		sizeof(context->chap_name));
-	strncpy((char *)rec->session.auth.password, context->chap_password,
+	strlcpy((char *)rec->session.auth.password, context->chap_password,
 		sizeof(context->chap_password));
-	strncpy(rec->session.auth.username_in, context->chap_name_in,
+	strlcpy(rec->session.auth.username_in, context->chap_name_in,
 		sizeof(context->chap_name_in));
-	strncpy((char *)rec->session.auth.password_in,
+	strlcpy((char *)rec->session.auth.password_in,
 		context->chap_password_in,
 		sizeof(context->chap_password_in));
 	rec->session.auth.password_length =
@@ -1958,35 +1617,83 @@ fw_create_rec_by_entry(struct boot_context *context)
 	return rec;
 }
 
-static int exec_fw_op(discovery_rec_t *drec, int do_login, int info_level)
+static int exec_fw_op(discovery_rec_t *drec, struct list_head *ifaces,
+		      int info_level, int do_login, int op)
 {
-	struct boot_context context;
-	struct node_rec *rec;
-	int ret = 0;
+	struct boot_context *context;
+	struct list_head targets, rec_list;
+	struct iface_rec *iface, *tmp_iface;
+	struct node_rec *rec, *tmp_rec;
+	int rc = 0;
 
-	memset(&context, 0, sizeof(struct boot_context));
-	ret = fw_get_entry(&context, NULL);
-	if (ret) {
-		log_error("Could not read fw values.");
-		return ret;
+	INIT_LIST_HEAD(&targets);
+	INIT_LIST_HEAD(&rec_list);
+
+	if (drec) {
+		/*
+		 * compat: if the user did not pass any op then we do all
+		 * ops for them
+		 */
+		if (!op)
+			op = OP_NEW | OP_DELETE | OP_UPDATE;
+
+		list_for_each_entry_safe(iface, tmp_iface, ifaces, list) {
+			rc = iface_conf_read(iface);
+			if (rc) {
+				log_error("Could not read iface info for %s. "
+					  "Make sure a iface config with the "
+					  "file name and iface.iscsi_ifacename "
+					  "%s is in %s.", iface->name,
+					  iface->name, IFACE_CONFIG_DIR);
+				list_del_init(&iface->list);
+				free(iface);
+				continue;
+			}
+		}
+
+		rc = idbm_bind_ifaces_to_nodes(discovery_fw, drec,
+					       ifaces, &rec_list);
+		if (rc)
+			log_error("Could not perform fw discovery.\n");
+		else
+			rc = exec_disc_op_on_recs(drec, &rec_list, info_level,
+						   do_login, op);
+
+		list_for_each_entry_safe(rec, tmp_rec, &rec_list, list) {
+			list_del(&rec->list);
+			free(rec);
+		}
+		return rc;
 	}
 
-	rec = fw_create_rec_by_entry(&context);
-	if (!rec)
-		return ENODEV;
+	/* The following ops do not interact with the DB */
+	rc = fw_get_targets(&targets);
+	if (rc) {
+		log_error("Could not get list of targets from firmware. "
+			  "(err %d)\n", rc);
+		return rc;
+	}
 
-	/* if discovery, print nodes that were found. */
-	if (drec)
-		print_fw_nodes(rec, info_level);
+	if (do_login) {
+		list_for_each_entry(context, &targets, list) {
+			rec = fw_create_rec_by_entry(context);
+			if (!rec) {
+				log_error("Could not convert firmware info to "
+					  "node record.\n");
+				rc = ENOMEM;
+				break;
+			}
 
-	if (do_login)
-		ret = login_portal(NULL, NULL, rec);
-	free(rec);
+			login_portal(NULL, NULL, rec);
+			free(rec);
+		}
+	} else {
+		list_for_each_entry(context, &targets, list)
+			fw_print_entry(context);
+	}
 
-	/* print the fw node info if called in fw mode with no params */
-	if (!do_login && !drec)
-		fw_print_entry(&context);
-	return ret;
+	fw_free_targets(&targets);
+	return rc;
 }
 
 int
@@ -2004,7 +1711,9 @@ main(int argc, char **argv)
 	struct list_head ifaces;
 	struct iface_rec *iface = NULL, *tmp;
 	struct node_rec *rec = NULL;
+	uint32_t host_no = -1;
 
+	memset(&drec, 0, sizeof(discovery_rec_t));
 	INIT_LIST_HEAD(&ifaces);
 	/* do not allow ctrl-c for now... */
 	memset(&sa_old, 0, sizeof(struct sigaction));
@@ -2054,6 +1763,16 @@ main(int argc, char **argv)
 		case 'v':
 			value = optarg;
 			break;
+		case 'H':
+			errno = 0;
+			host_no = strtoul(optarg, NULL, 10);
+			if (errno) {
+				log_error("invalid host no %s. %s.",
+					  optarg, strerror(errno));
+				rc = -1;
+				goto free_ifaces;
+			}
+			break;
 		case 'r':
 			sid = iscsi_sysfs_get_sid_from_path(optarg);
 			if (sid < 0) {
@@ -2149,7 +1868,7 @@ main(int argc, char **argv)
 			goto free_ifaces;
 		}
 
-		rc = exec_fw_op(NULL, do_login, info_level);
+		rc = exec_fw_op(NULL, NULL, info_level, do_login, op);
 		goto free_ifaces;
 	}
 
@@ -2163,6 +1882,16 @@ main(int argc, char **argv)
 	iface_setup_host_bindings();
 
 	switch (mode) {
+	case MODE_HOST:
+		if ((rc = verify_mode_params(argc, argv, "HdmP", 0))) {
+			log_error("host mode: option '-%c' is not "
+				  "allowed/supported", rc);
+			rc = -1;
+			goto out;
+		}
+
+		rc = host_info_print(info_level, host_no);
+		break;
 	case MODE_IFACE:
 		if ((rc = verify_mode_params(argc, argv, "IdnvmPo", 0))) {
 			log_error("iface mode: option '-%c' is not "
@@ -2183,7 +1912,7 @@ main(int argc, char **argv)
 				   name, value);
 		break;
 	case MODE_DISCOVERY:
-		if ((rc = verify_mode_params(argc, argv, "IPdmtplo", 0))) {
+		if ((rc = verify_mode_params(argc, argv, "SIPdmtplo", 0))) {
 			log_error("discovery mode: option '-%c' is not "
 				  "allowed/supported", rc);
 			rc = -1;
@@ -2198,9 +1927,8 @@ main(int argc, char **argv)
 				goto out;
 			}
 
-			memset(&drec, 0, sizeof(discovery_rec_t));
 			idbm_sendtargets_defaults(&drec.u.sendtargets);
-			strncpy(drec.address, ip, sizeof(drec.address));
+			strlcpy(drec.address, ip, sizeof(drec.address));
 			drec.port = port;
 
 			if (do_sendtargets(&drec, &ifaces, info_level,
@@ -2220,9 +1948,10 @@ main(int argc, char **argv)
 			if (isns_dev_attr_query(&drec, info_level))
 				rc = -1;
 			break;
-		case DISCOVERY_TYPE_FWBOOT:
-			drec.type = DISCOVERY_TYPE_FWBOOT;
-			if (exec_fw_op(&drec, do_login, info_level))
+		case DISCOVERY_TYPE_FW:
+			drec.type = DISCOVERY_TYPE_FW;
+			if (exec_fw_op(&drec, &ifaces, info_level, do_login,
+					op))
 				rc = -1;
 			break;
 		default:
@@ -2364,7 +2093,7 @@ main(int argc, char **argv)
 
 			if (!do_logout && !do_rescan && !do_stats &&
 			    op == OP_NOOP && info_level > 0) {
-				rc = print_sessions(info_level, info);
+				rc = session_info_print(info_level, info);
 				if (rc)
 					rc = -1;
 				goto free_info;
@@ -2395,7 +2124,7 @@ free_info:
 				goto out;
 			}
 
-			rc = print_sessions(info_level, NULL);
+			rc = session_info_print(info_level, NULL);
 		}
 		break;
 	default:
