@@ -34,6 +34,7 @@
 #include "log.h"
 #include "util.h"
 #include "isns_proto.h"
+#include "sysdeps.h"
 
 enum isns_task_state {
 	ISNS_TASK_WAIT_CONN,
@@ -50,16 +51,17 @@ struct isns_task {
 	int done;
 	int retry;
 	queue_task_t *qtask;
+	struct actor actor;
 };
 
-static actor_t isns_actor;
-static queue_t *isns_queue = NULL;
 static struct sockaddr_storage ss;
 static uint16_t transaction;
 
 static char isns_address[NI_MAXHOST];
 static int isns_port = 3205, isns_listen_port, max_retry = 10000;
 
+static void isns_poll(void *data);
+
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 
 #define get_hdr_param(hdr, function, length, flags, transaction, sequence)	\
@@ -270,17 +272,24 @@ static int isns_recv_pdu(struct isns_task *task)
 	return 0;
 }
 
+static char *isns_get_config_file(void)
+{
+	return dconfig->config_file;
+}
+
 static void add_new_target_node(char *targetname, uint8_t *ip, int port,
 				int tag)
 {
 	int err;
 	node_rec_t rec;
 	discovery_rec_t drec;
-	idbm_t *db;
 	char dst[INET6_ADDRSTRLEN];
 
 	memset(dst, 0, sizeof(dst));
-	if (!memcmp(ip, dst, 10) && ip[10] == 0xff && ip[11] == 0xff)
+	/*
+	 * some servers are sending compat instead of mapped
+	 */
+	if (IN6_IS_ADDR_V4MAPPED(ip) || IN6_IS_ADDR_V4COMPAT(ip))
 		inet_ntop(AF_INET, ip + 12, dst, sizeof(dst));
 	else
 		inet_ntop(AF_INET6, ip, dst, sizeof(dst));
@@ -288,26 +297,26 @@ static void add_new_target_node(char *targetname, uint8_t *ip, int port,
 	log_debug(1, "add a new target node:%s %s,%d %d",
 		  targetname, dst, port, tag);
 
-	db = idbm_init(dconfig->config_file);
-	if (!db) {
+	if (idbm_init(isns_get_config_file)) {
 		log_error("Could not add new target node:%s %s,%d",
 			  targetname, dst, port);
 		return;
 	}
-	idbm_node_setup_from_conf(db, &rec);
-	strncpy(rec.name, targetname, TARGET_NAME_MAXLEN);
+	idbm_node_setup_from_conf(&rec);
+	strlcpy(rec.name, targetname, TARGET_NAME_MAXLEN);
 	rec.conn[0].port = port;
 	rec.tpgt = tag;
-	strncpy(rec.conn[0].address, dst, NI_MAXHOST);
+	strlcpy(rec.conn[0].address, dst, NI_MAXHOST);
 
 	/* TODO?: shoudl we set the address and port of the server ? */
+	memset(&drec, 0, sizeof(discovery_rec_t));
 	drec.type = DISCOVERY_TYPE_ISNS;
-	err = idbm_add_nodes(db, &rec, &drec, NULL);
+	err = idbm_add_nodes(&rec, &drec, NULL, 0);
 	if (err)
 		log_error("Could not add new target node:%s %s,%d",
 			  targetname, dst, port);
 
-	idbm_terminate(db);
+	idbm_terminate();
 }
 
 static int qry_rsp_handle(struct isns_hdr *hdr)
@@ -432,8 +441,8 @@ static int isns_task_done(struct isns_task *task)
 		task->len = length + sizeof(*hdr);
 		task->done = 0;
 
-		queue_produce(isns_queue, EV_CONN_POLL, task, 0, NULL);
-		actor_schedule(&isns_actor);
+		actor_new(&task->actor, isns_poll, task);
+		actor_schedule(&task->actor);
 		finished = 0;
 		break;
 	default:
@@ -502,8 +511,8 @@ int isns_dev_attr_query_task(queue_task_t *qtask)
 
 	qtask->rsp.command = MGMT_IPC_ISNS_DEV_ATTR_QUERY;
 
-	queue_produce(isns_queue, EV_CONN_POLL, task, 0, NULL);
-	actor_schedule(&isns_actor);
+	actor_new(&task->actor, isns_poll, task);
+	actor_schedule(&task->actor);
 
 	return MGMT_IPC_OK;
 }
@@ -532,15 +541,15 @@ void isns_handle(int listen_fd)
 	task->state = ISNS_TASK_RECV_PDU;
 	task->fd = fd;
 
-	queue_produce(isns_queue, EV_CONN_POLL, task, 0, NULL);
-	actor_schedule(&isns_actor);
+	actor_new(&task->actor, isns_poll, task);
+	actor_schedule(&task->actor);
 }
 
-static void isns_poll(queue_item_t *item)
+static void isns_poll(void *data)
 {
 	int err, finished;
 	struct pollfd pfd;
-	struct isns_task *task = item->context;
+	struct isns_task *task = data;
 	struct isns_hdr *hdr = (struct isns_hdr *) task->data;
 	uint16_t function = ntohs(hdr->function);
 
@@ -573,9 +582,8 @@ static void isns_poll(queue_item_t *item)
 						goto free_task;
 				}
 
-				queue_produce(isns_queue, EV_CONN_POLL, task, 0,
-					      NULL);
-				actor_schedule(&isns_actor);
+				actor_new(&task->actor, isns_poll, task);
+				actor_schedule(&task->actor);
 			}
 			break;
 		case ISNS_TASK_RECV_PDU:
@@ -590,9 +598,9 @@ static void isns_poll(queue_item_t *item)
 						goto free_task;
 				} else {
 					/* need to read more */
-					queue_produce(isns_queue, EV_CONN_POLL,
-						      task, 0, NULL);
-					actor_schedule(&isns_actor);
+					actor_new(&task->actor, isns_poll,
+						  task);
+					actor_schedule(&task->actor);
 				}
 			}
 		}
@@ -602,8 +610,8 @@ static void isns_poll(queue_item_t *item)
 			log_error("abort task");
 			goto abort_task;
 		} else {
-			queue_produce(isns_queue, EV_CONN_POLL, task, 0, NULL);
-			actor_schedule(&isns_actor);
+			actor_new(&task->actor, isns_poll, task);
+			actor_schedule(&task->actor);
 		}
 	}
 
@@ -615,32 +623,6 @@ free_task:
 	isns_free_task(task);
 }
 
-static void isns_control(void *data)
-{
-	int count = isns_queue->count, i;
-	int err;
-	unsigned char item_buf[sizeof(queue_item_t) + EVENT_PAYLOAD_MAX];
-	queue_item_t *item = (queue_item_t *)(void *)item_buf;
-
-	for (i = 0; i < count; i++) {
-		err = queue_consume(isns_queue, EVENT_PAYLOAD_MAX, item);
-		if (err == QUEUE_IS_EMPTY) {
-			log_debug(4, "%d items flushed while mainloop "
-				  "was processing", count - i);
-			break;
-		}
-
-		switch (item->event_type) {
-		case EV_CONN_POLL:
-			isns_poll(item);
-			break;
-		default:
-			log_error("%d unknown event type", item->event_type);
-			break;
-		}
-	}
-}
-
 static int isns_dev_register(void)
 {
 	struct isns_task *task;
@@ -658,10 +640,9 @@ static int isns_dev_register(void)
 
 	task->state = ISNS_TASK_WAIT_CONN;
 	build_dev_reg_req(task);
-	queue_produce(isns_queue, EV_CONN_POLL, task, 0, NULL);
 
-	actor_new(&isns_actor, isns_control, NULL);
-	actor_schedule(&isns_actor);
+	actor_new(&task->actor, isns_poll, task);
+	actor_schedule(&task->actor);
 
 	return 0;
 }
@@ -722,7 +703,7 @@ int isns_init(void)
 	int fd = -1, err;
 	FILE *f;
 
-	f = fopen(dconfig->config_file, "r");
+	f = fopen(isns_get_config_file(), "r");
 	if (!f)
 		return -EIO;
 
@@ -741,62 +722,23 @@ int isns_init(void)
 	if (!strlen(isns_address))
 		return -1;
 
-	isns_queue = queue_create(4, 4, NULL, NULL);
-	if (!isns_queue) {
-		log_error("can't create queue %m");
-		return -ENOMEM;
-	}
-
 	snprintf(port, sizeof(port), "%d", isns_port);
 	err = resolve_address(isns_address, port, &ss);
 	if (err) {
 		log_error("can't resolve address %m, %s", isns_address);
-		goto free_queue;
+		return err;
 	}
 
 	err = isns_listen_init(&fd);
 	if (err)
-		goto free_queue;
+		return err;
 
 	isns_dev_register();
 	return fd;
-
-free_queue:
-	queue_destroy(isns_queue);
-	return err;
 }
 
 void isns_exit(void)
 {
-	int err, count, i;
-	unsigned char item_buf[sizeof(queue_item_t) + EVENT_PAYLOAD_MAX];
-	queue_item_t *item = (queue_item_t *)(void *)item_buf;
-
-	if (!isns_queue)
-		return;
-
-	count = isns_queue->count;
-	/*
-	 * TODO: Add some code to gracefully shutdown.
-	 */
-	for (i = 0; i < count; i++) {
-		err = queue_consume(isns_queue, EVENT_PAYLOAD_MAX, item);
-		if (err == QUEUE_IS_EMPTY) {
-			log_debug(4, "%d items flushed while mainloop "
-				  "was processing", count - i);
-			break;
-		}
-
-		log_debug(4, "Dropping event type %d\n", item->event_type);
-		switch (item->event_type) {
-		case EV_CONN_POLL:
-			isns_free_task(item->context);
-			continue;
-		default:
-			log_error("%d unknown event type", item->event_type);
-			continue;
-		}
-	}
-
-	queue_destroy(isns_queue);
+	/* do nothing for now */
+	;
 }
